"""Tests for subarulink connection functions."""
import asyncio
import time

import pytest

import subarulink
from subarulink._subaru_api.const import (
    API_2FA_AUTH_VERIFY,
    API_2FA_CONTACT,
    API_2FA_SEND_VERIFICATION,
    API_CONDITION,
    API_G2_FETCH_RES_SUBARU_PRESETS,
    API_G2_FETCH_RES_USER_PRESETS,
    API_G2_LOCATE_STATUS,
    API_G2_LOCATE_UPDATE,
    API_HORN_LIGHTS,
    API_LIGHTS,
    API_LOCATE,
    API_LOGIN,
    API_REMOTE_SVC_STATUS,
    API_SELECT_VEHICLE,
    API_VALIDATE_SESSION,
    API_VEHICLE_HEALTH,
    API_VEHICLE_STATUS,
)
from subarulink.const import FETCH_INTERVAL, POLL_INTERVAL
from subarulink.exceptions import SubaruException

from tests.api_responses import (
    ERROR_403,
    ERROR_VEHICLE_SETUP,
    ERROR_VIN_NOT_FOUND,
    FETCH_SUBARU_CLIMATE_PRESETS,
    FETCH_USER_CLIMATE_PRESETS_EV,
    LOCATE_G2,
    LOGIN_ERRORS,
    LOGIN_INVALID_PASSWORD,
    LOGIN_MULTI_REGISTERED,
    LOGIN_SINGLE_NOT_REGISTERED,
    LOGIN_SINGLE_REGISTERED,
    REMOTE_CMD_INVALID_PIN,
    REMOTE_SERVICE_EXECUTE,
    REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
    REMOTE_SERVICE_STATUS_STARTED,
    SELECT_VEHICLE_1,
    SELECT_VEHICLE_2,
    SELECT_VEHICLE_3,
    VALIDATE_SESSION_FAIL,
    VALIDATE_SESSION_SUCCESS,
    VEHICLE_CONDITION_EV,
    VEHICLE_HEALTH_EV,
    VEHICLE_STATUS_EV,
    VEHICLE_STATUS_EXECUTE,
    VEHICLE_STATUS_FINISHED_SUCCESS,
    VEHICLE_STATUS_STARTED,
)
from tests.conftest import (
    TEST_DEVICE_ID,
    TEST_DEVICE_NAME,
    TEST_PASSWORD,
    TEST_PIN,
    TEST_USERNAME,
    TEST_VIN_1_G1,
    TEST_VIN_2_EV,
    TEST_VIN_3_G2,
    TEST_VIN_4_SAFETY_PLUS,
    TEST_VIN_5_G1_SECURITY,
    server_js_response,
)


async def test_connect_incomplete_credentials():
    controller = subarulink.Controller(
        None,
        TEST_USERNAME,
        None,
        TEST_DEVICE_ID,
        TEST_PIN,
        TEST_DEVICE_NAME,
    )
    task = asyncio.create_task(controller.connect())

    with pytest.raises(subarulink.SubaruException):
        await task


async def test_no_dns(http_redirect):
    controller = subarulink.Controller(
        http_redirect.session,
        TEST_USERNAME,
        TEST_PASSWORD,
        TEST_DEVICE_ID,
        TEST_PIN,
        TEST_DEVICE_NAME,
    )
    task = asyncio.create_task(controller.connect())

    with pytest.raises(subarulink.SubaruException):
        await task


async def test_connect_fail_authenticate(test_server, controller):
    task = asyncio.create_task(controller.connect())

    await server_js_response(test_server, LOGIN_INVALID_PASSWORD, path=API_LOGIN)
    with pytest.raises(subarulink.SubaruException):
        await task


async def test_handle_404(test_server, controller):
    task = asyncio.create_task(controller.connect())

    await server_js_response(test_server, LOGIN_INVALID_PASSWORD, path=API_LOGIN, status=404)
    with pytest.raises(subarulink.SubaruException):
        await task


async def test_connect_device_registration_success(test_server, controller):
    task = asyncio.create_task(controller.connect())

    await server_js_response(test_server, LOGIN_SINGLE_NOT_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_1,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_1_G1, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server, {"success": True, "data": {"userName": "test@test.com"}}, path=API_2FA_CONTACT
    )
    assert await task

    task = asyncio.create_task(controller.request_auth_code("userName"))
    await server_js_response(test_server, {"success": True}, path=API_2FA_SEND_VERIFICATION)
    assert await task

    task = asyncio.create_task(controller.submit_auth_code("123456"))
    await server_js_response(test_server, {"success": True}, path=API_2FA_AUTH_VERIFY)
    await server_js_response(test_server, LOGIN_SINGLE_NOT_REGISTERED, path=API_LOGIN)
    await server_js_response(test_server, LOGIN_SINGLE_NOT_REGISTERED, path=API_LOGIN)
    await server_js_response(test_server, LOGIN_SINGLE_REGISTERED, path=API_LOGIN)
    assert await task


async def test_connect_device_registration_bad_input(test_server, controller):
    task = asyncio.create_task(controller.connect())

    await server_js_response(test_server, LOGIN_SINGLE_NOT_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_1,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_1_G1, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server, {"success": True, "data": {"userName": "test@test.com"}}, path=API_2FA_CONTACT
    )
    assert await task

    task = asyncio.create_task(controller.request_auth_code("invalid_auth_type"))
    assert not await task

    task = asyncio.create_task(controller.submit_auth_code("invalid_code_format"))
    assert not await task


async def test_connect_single_car(single_vehicle_controller):
    assert single_vehicle_controller.get_vehicles() == [TEST_VIN_1_G1]
    assert single_vehicle_controller.get_ev_status(TEST_VIN_1_G1) is False


async def test_connect_multi_car(multi_vehicle_controller):
    vehicles = multi_vehicle_controller.get_vehicles()
    assert TEST_VIN_1_G1 in vehicles
    assert TEST_VIN_2_EV in vehicles
    assert TEST_VIN_3_G2 in vehicles
    assert TEST_VIN_4_SAFETY_PLUS in vehicles
    assert TEST_VIN_5_G1_SECURITY in vehicles
    assert not multi_vehicle_controller.get_ev_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_ev_status(TEST_VIN_2_EV)
    assert not multi_vehicle_controller.get_remote_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_remote_status(TEST_VIN_3_G2)
    assert not multi_vehicle_controller.get_res_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_remote_status(TEST_VIN_3_G2)
    assert not multi_vehicle_controller.get_safety_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_safety_status(TEST_VIN_4_SAFETY_PLUS)


async def test_login_fail(test_server, controller):
    for fail_msg in LOGIN_ERRORS:
        task = asyncio.create_task(controller.connect())
        await server_js_response(test_server, fail_msg, path=API_LOGIN)
        with pytest.raises(subarulink.SubaruException):
            await task


async def test_test_pin_success(test_server, multi_vehicle_controller):
    assert multi_vehicle_controller.is_pin_required()
    task = asyncio.create_task(multi_vehicle_controller.test_pin())
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, VEHICLE_STATUS_EV, path=API_G2_LOCATE_UPDATE)
    assert await task


async def test_test_pin_fail(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.test_pin())
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, REMOTE_CMD_INVALID_PIN, path=API_G2_LOCATE_UPDATE)
    with pytest.raises(subarulink.InvalidPIN):
        await task

    assert not multi_vehicle_controller.update_saved_pin(TEST_PIN)
    assert multi_vehicle_controller.update_saved_pin("0000")


async def test_test_pin_not_needed(single_vehicle_controller):
    assert not single_vehicle_controller.is_pin_required()
    task = asyncio.create_task(single_vehicle_controller.test_pin())
    assert not await task


async def test_switch_vehicle_success(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_2_EV))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, REMOTE_SERVICE_EXECUTE, path=API_LIGHTS)
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=API_REMOTE_SVC_STATUS,
    )

    assert await task


async def test_switch_vehicle_fail(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_2_EV))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        ERROR_VIN_NOT_FOUND,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    with pytest.raises(SubaruException):
        await task


async def test_switch_vehicle_setup_fail(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.horn(TEST_VIN_2_EV))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        ERROR_VEHICLE_SETUP,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, LOGIN_SINGLE_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_HORN_LIGHTS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=API_REMOTE_SVC_STATUS,
    )
    assert await task


async def test_expired_session(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.horn(TEST_VIN_3_G2))

    await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_HORN_LIGHTS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=API_REMOTE_SVC_STATUS,
    )

    assert await task


async def test_403_during_remote_query(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.get_data(TEST_VIN_2_EV))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, VEHICLE_STATUS_EV, path=API_VEHICLE_STATUS)
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, ERROR_403, path=API_CONDITION)
    await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)

    await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, VEHICLE_CONDITION_EV, path=API_CONDITION)
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, LOCATE_G2, path=API_LOCATE)
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, VEHICLE_HEALTH_EV, path=API_VEHICLE_HEALTH)
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)

    await task


async def test_403_during_remote_command(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.update(TEST_VIN_2_EV))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        ERROR_403,
        path=API_G2_LOCATE_UPDATE,
    )

    # 403 Handling and reattempt
    await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        VEHICLE_STATUS_EXECUTE,
        path=API_G2_LOCATE_UPDATE,
    )

    # Back to normal
    await server_js_response(test_server, VEHICLE_STATUS_STARTED, path=API_G2_LOCATE_STATUS)
    await server_js_response(
        test_server,
        VEHICLE_STATUS_FINISHED_SUCCESS,
        path=API_G2_LOCATE_STATUS,
    )

    assert await task


async def test_interval_functions(multi_vehicle_controller):
    INVALID_NEW_INTERVAL = 25
    VALID_NEW_INTERVAL = 500

    assert multi_vehicle_controller.get_update_interval() == POLL_INTERVAL
    multi_vehicle_controller.set_update_interval(INVALID_NEW_INTERVAL)
    assert multi_vehicle_controller.get_update_interval() == POLL_INTERVAL
    multi_vehicle_controller.set_update_interval(VALID_NEW_INTERVAL)
    assert multi_vehicle_controller.get_update_interval() == VALID_NEW_INTERVAL
    assert multi_vehicle_controller.get_fetch_interval() == FETCH_INTERVAL
    multi_vehicle_controller.set_fetch_interval(INVALID_NEW_INTERVAL)
    assert multi_vehicle_controller.get_fetch_interval() == FETCH_INTERVAL
    multi_vehicle_controller.set_fetch_interval(VALID_NEW_INTERVAL)
    assert multi_vehicle_controller.get_fetch_interval() == VALID_NEW_INTERVAL



#  SPDX-License-Identifier: Apache-2.0
"""
Constants used by subarulink to interact with the STARLINK HTTP API.

This is an undocumented API derived from analysis of the MySubaru Android app v2.7.30.
"""



#  SPDX-License-Identifier: Apache-2.0
"""
An example console application that uses the subarulink package.

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""



#  SPDX-License-Identifier: Apache-2.0
"""
Constants used by subarulink to interact with the STARLINK HTTP API.

This is an undocumented API derived from analysis of the MySubaru Android app v2.7.39.
"""

API_VERSION = "/g2v27"

API_SERVER = {
    "USA": "mobileapi.prod.subarucs.com",
    "CAN": "mobileapi.ca.prod.subarucs.com",
}

API_MOBILE_APP = {
    "USA": "com.subaru.telematics.app.remote",
    "CAN": "ca.subaru.telematics.remote",
}

# Same API for g1 and g2
API_LOGIN = "/login.json"
API_2FA_CONTACT = "/twoStepAuthContacts.json"
API_2FA_SEND_VERIFICATION = "/twoStepAuthSendVerification.json"
API_2FA_AUTH_VERIFY = "/twoStepAuthVerify.json"
API_REFRESH_VEHICLES = "/refreshVehicles.json"
API_SELECT_VEHICLE = "/selectVehicle.json"
API_VALIDATE_SESSION = "/validateSession.json"
API_VEHICLE_STATUS = "/vehicleStatus.json"
API_AUTHORIZE_DEVICE = "/authenticateDevice.json"
API_NAME_DEVICE = "/nameThisDevice.json"
API_VEHICLE_HEALTH = "/vehicleHealth.json"

# Similar API for g1 and g2 -- controller must replace 'api_gen' with either 'g1' or 'g2'
API_LOCK = "/service/api_gen/lock/execute.json"
API_LOCK_CANCEL = "/service/api_gen/lock/cancel.json"

API_UNLOCK = "/service/api_gen/unlock/execute.json"
API_UNLOCK_CANCEL = "/service/api_gen/unlock/cancel.json"

API_HORN_LIGHTS = "/service/api_gen/hornLights/execute.json"
API_HORN_LIGHTS_CANCEL = "/service/api_gen/hornLights/cancel.json"
API_HORN_LIGHTS_STOP = "/service/api_gen/hornLights/stop.json"

API_LIGHTS = "/service/api_gen/lightsOnly/execute.json"
API_LIGHTS_CANCEL = "/service/api_gen/lightsOnly/cancel.json"
API_LIGHTS_STOP = "/service/api_gen/lightsOnly/stop.json"

API_CONDITION = "/service/api_gen/condition/execute.json"
API_LOCATE = "/service/api_gen/locate/execute.json"
API_REMOTE_SVC_STATUS = "/service/api_gen/remoteService/status.json"

# Different API for g1 and g2
API_G1_LOCATE_UPDATE = "/service/g1/vehicleLocate/execute.json"
API_G1_LOCATE_STATUS = "/service/g1/vehicleLocate/status.json"
API_G2_LOCATE_UPDATE = "/service/g2/vehicleStatus/execute.json"
API_G2_LOCATE_STATUS = "/service/g2/vehicleStatus/locationStatus.json"

# g1-Only API
API_G1_HORN_LIGHTS_STATUS = "/service/g1/hornLights/status.json"

# g2-Only API
API_G2_SEND_POI = "/service/g2/sendPoi/execute.json"
API_G2_SPEEDFENCE = "/service/g2/speedFence/execute.json"
API_G2_GEOFENCE = "/service/g2/geoFence/execute.json"
API_G2_CURFEW = "/service/g2/curfew/execute.json"

API_G2_REMOTE_ENGINE_START = "/service/g2/engineStart/execute.json"
API_G2_REMOTE_ENGINE_START_CANCEL = "/service/g2/engineStart/cancel.json"
API_G2_REMOTE_ENGINE_STOP = "/service/g2/engineStop/execute.json"

API_G2_FETCH_RES_QUICK_START_SETTINGS = "/service/g2/remoteEngineQuickStartSettings/fetch.json"
API_G2_FETCH_RES_USER_PRESETS = "/service/g2/remoteEngineStartSettings/fetch.json"
API_G2_FETCH_RES_SUBARU_PRESETS = "/service/g2/climatePresetSettings/fetch.json"
API_G2_SAVE_RES_SETTINGS = "/service/g2/remoteEngineStartSettings/save.json"
API_G2_SAVE_RES_QUICK_START_SETTINGS = "/service/g2/remoteEngineQuickStartSettings/save.json"

# EV-Only API
API_EV_CHARGE_NOW = "/service/g2/phevChargeNow/execute.json"
API_EV_FETCH_CHARGE_SETTINGS = "/service/g2/phevGetTimerSettings/execute.json"
API_EV_SAVE_CHARGE_SETTINGS = "/service/g2/phevSendTimerSetting/execute.json"
API_EV_DELETE_CHARGE_SCHEDULE = "/service/g2/phevDeleteTimerSetting/execute.json"


# vehicleStatus.json fields
API_AVG_FUEL_CONSUMPTION = "avgFuelConsumptionLitersPer100Kilometers"
API_DIST_TO_EMPTY = "distanceToEmptyFuelKilometers"
API_TIMESTAMP = "eventDateStr"
API_LATITUDE = "latitude"
API_LONGITUDE = "longitude"
API_HEADING = "positionHeadingDegree"
API_ODOMETER = "odometerValueKilometers"
API_VEHICLE_STATE = "vehicleStateType"
API_TIRE_PRESSURE_FL = "tirePressureFrontLeft"
API_TIRE_PRESSURE_FR = "tirePressureFrontRight"
API_TIRE_PRESSURE_RL = "tirePressureRearLeft"
API_TIRE_PRESSURE_RR = "tirePressureRearRight"

# vehicleHealth.json fields
API_HEALTH_TROUBLE = "isTrouble"
API_HEALTH_ONDATES = "onDates"
API_HEALTH_FEATURE = "featureCode"

# condition/execute.json fields
API_DOOR_BOOT_POSITION = "doorBootPosition"
API_DOOR_ENGINE_HOOD_POSITION = "doorEngineHoodPosition"
API_DOOR_FRONT_LEFT_POSITION = "doorFrontLeftPosition"
API_DOOR_FRONT_RIGHT_POSITION = "doorFrontRightPosition"
API_DOOR_REAR_LEFT_POSITION = "doorRearLeftPosition"
API_DOOR_REAR_RIGHT_POSITION = "doorRearRightPosition"
API_EV_CHARGER_STATE_TYPE = "evChargerStateType"
API_EV_DISTANCE_TO_EMPTY = "evDistanceToEmpty"
API_EV_IS_PLUGGED_IN = "evIsPluggedIn"
API_EV_STATE_OF_CHARGE_MODE = "evStateOfChargeMode"
API_EV_STATE_OF_CHARGE_PERCENT = "evStateOfChargePercent"
API_EV_TIME_TO_FULLY_CHARGED = "evTimeToFullyCharged"
API_EV_TIME_TO_FULLY_CHARGED_UTC = "evTimeToFullyChargedUTC"
API_REMAINING_FUEL_PERCENT = "remainingFuelPercent"
API_LAST_UPDATED_DATE = "lastUpdatedTime"
API_WINDOW_FRONT_LEFT_STATUS = "windowFrontLeftStatus"
API_WINDOW_FRONT_RIGHT_STATUS = "windowFrontRightStatus"
API_WINDOW_REAR_LEFT_STATUS = "windowRearLeftStatus"
API_WINDOW_REAR_RIGHT_STATUS = "windowRearRightStatus"
API_WINDOW_SUNROOF_STATUS = "windowSunroofStatus"

# Timestamp formats
API_TIMESTAMP_FMT = "%Y-%m-%dT%H:%M:%S%z"  # "2020-04-25T23:35:55+0000"
API_VS_TIMESTAMP_FMT = "%Y-%m-%dT%H:%M%z"  # "2020-04-25T23:35+0000"
API_POSITION_TIMESTAMP_FMT = "%Y-%m-%dT%H:%M:%SZ"  # "2020-04-25T23:35:55Z"

# selectVehicle.json keys
API_VEHICLE_ATTRIBUTES = "attributes"
API_VEHICLE_ID = "id"
API_VEHICLE_MODEL_NAME = "modelName"
API_VEHICLE_MODEL_YEAR = "modelYear"
API_VEHICLE_NAME = "nickname"
API_VEHICLE_API_GEN = "api_gen"
API_VEHICLE_FEATURES = "features"
API_VEHICLE_SUBSCRIPTION_FEATURES = "subscriptionFeatures"
API_VEHICLE_SUBSCRIPTION_STATUS = "subscriptionStatus"

# API_VEHICLE_FEATURES items that determine available functionality
API_FEATURE_PHEV = "PHEV"
API_FEATURE_REMOTE_START = "RES"
API_FEATURE_REMOTE = "REMOTE"
API_FEATURE_SAFETY = "SAFETY"
API_FEATURE_ACTIVE = "ACTIVE"
API_FEATURE_PHEV = "PHEV"
API_FEATURE_REMOTE_START = "RES"
API_FEATURE_MOONROOF_PANORAMIC = "PANPM-DG2G"
API_FEATURE_MOONROOF_POWER = "PANPM-TUIRWAOC"
API_FEATURE_POWER_WINDOWS = "PWAAADWWAP"
API_FEATURE_FRONT_TIRE_RECOMMENDED_PRESSURE_PREFIX = "TIF_"
API_FEATURE_REAR_TIRE_RECOMMENDED_PRESSURE_PREFIX = "TIR_"
API_FEATURE_G1_TELEMATICS = "g1"
API_FEATURE_G2_TELEMATICS = "g2"
API_FEATURE_G3_TELEMATICS = "g3"

# G2 Error Codes
API_ERROR_SOA_403 = "403-soa-unableToParseResponseBody"
API_ERROR_INVALID_CREDENTIALS = "InvalidCredentials"
API_ERROR_SERVICE_ALREADY_STARTED = "ServiceAlreadyStarted"
API_ERROR_INVALID_ACCOUNT = "invalidAccount"
API_ERROR_PASSWORD_WARNING = "passwordWarning"
API_ERROR_ACCOUNT_LOCKED = "accountLocked"
API_ERROR_NO_VEHICLES = "noVehiclesOnAccount"
API_ERROR_NO_ACCOUNT = "accountNotFound"
API_ERROR_TOO_MANY_ATTEMPTS = "tooManyAttempts"
API_ERROR_VEHICLE_NOT_IN_ACCOUNT = "vehicleNotInAccount"
API_ERROR_INVALID_TOKEN = "InvalidToken"
API_ERROR_VEHICLE_SETUP = "VEHICLESETUPERROR"

# G1 Error Codes
API_ERROR_G1_NO_SUBSCRIPTION = "SXM40004"
API_ERROR_G1_STOLEN_VEHICLE = "SXM40005"
API_ERROR_G1_INVALID_PIN = "SXM40006"
API_ERROR_G1_SERVICE_ALREADY_STARTED = "SXM40009"
API_ERROR_G1_PIN_LOCKED = "SXM40017"


API_MAX_SESSION_AGE_MINS = 240
API_SERVICE_REQ_ID = "serviceRequestId"



"""Tests for subarulink vehicle status functions."""
import asyncio

import pytest

from subarulink._subaru_api.const import (
    API_AVG_FUEL_CONSUMPTION,
    API_DIST_TO_EMPTY,
    API_G1_LOCATE_STATUS,
    API_G1_LOCATE_UPDATE,
    API_G2_LOCATE_STATUS,
    API_G2_LOCATE_UPDATE,
    API_LATITUDE,
    API_LOCATE,
    API_LONGITUDE,
    API_TIRE_PRESSURE_FL,
    API_TIRE_PRESSURE_FR,
    API_TIRE_PRESSURE_RL,
    API_TIRE_PRESSURE_RR,
    API_VEHICLE_STATE,
    API_VEHICLE_STATUS,
)
import subarulink.const as sc

from tests.api_responses import (
    LOCATE_G1_EXECUTE,
    LOCATE_G1_FINISHED,
    LOCATE_G1_STARTED,
    LOCATE_G2_BAD_LOCATION,
    SELECT_VEHICLE_1,
    SELECT_VEHICLE_2,
    SELECT_VEHICLE_3,
    SELECT_VEHICLE_4,
    SELECT_VEHICLE_5,
    VEHICLE_CONDITION_EV,
    VEHICLE_STATUS_EV,
    VEHICLE_STATUS_EV_MISSING_DATA,
    VEHICLE_STATUS_EXECUTE,
    VEHICLE_STATUS_FINISHED_SUCCESS,
    VEHICLE_STATUS_STARTED,
)
from tests.conftest import (
    TEST_VIN_1_G1,
    TEST_VIN_2_EV,
    TEST_VIN_3_G2,
    TEST_VIN_4_SAFETY_PLUS,
    TEST_VIN_5_G1_SECURITY,
    add_ev_vehicle_condition,
    add_ev_vehicle_status,
    add_fetch_climate_presets,
    add_g2_vehicle_locate,
    add_select_vehicle_sequence,
    add_validate_session,
    add_vehicle_health,
    assert_vehicle_condition,
    assert_vehicle_status,
    server_js_response,
)


async def test_vehicle_attributes(multi_vehicle_controller):
    assert multi_vehicle_controller.vin_to_name(TEST_VIN_1_G1) == SELECT_VEHICLE_1["data"]["nickname"]
    assert multi_vehicle_controller.vin_to_name(TEST_VIN_2_EV) == SELECT_VEHICLE_2["data"]["nickname"]
    assert multi_vehicle_controller.vin_to_name(TEST_VIN_3_G2) == SELECT_VEHICLE_3["data"]["nickname"]
    assert multi_vehicle_controller.vin_to_name(TEST_VIN_4_SAFETY_PLUS) == SELECT_VEHICLE_4["data"]["nickname"]
    assert multi_vehicle_controller.vin_to_name(TEST_VIN_5_G1_SECURITY) == SELECT_VEHICLE_5["data"]["nickname"]

    assert multi_vehicle_controller.get_api_gen(TEST_VIN_1_G1) == sc.FEATURE_G1_TELEMATICS
    assert multi_vehicle_controller.get_api_gen(TEST_VIN_2_EV) == sc.FEATURE_G2_TELEMATICS
    assert multi_vehicle_controller.get_api_gen(TEST_VIN_3_G2) == sc.FEATURE_G2_TELEMATICS
    assert multi_vehicle_controller.get_api_gen(TEST_VIN_4_SAFETY_PLUS) == sc.FEATURE_G2_TELEMATICS
    assert multi_vehicle_controller.get_api_gen(TEST_VIN_5_G1_SECURITY) == sc.FEATURE_G1_TELEMATICS

    assert not multi_vehicle_controller.get_safety_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_safety_status(TEST_VIN_2_EV)
    assert multi_vehicle_controller.get_safety_status(TEST_VIN_3_G2)
    assert multi_vehicle_controller.get_safety_status(TEST_VIN_4_SAFETY_PLUS)
    assert multi_vehicle_controller.get_safety_status(TEST_VIN_5_G1_SECURITY)

    assert not multi_vehicle_controller.get_ev_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_ev_status(TEST_VIN_2_EV)
    assert not multi_vehicle_controller.get_ev_status(TEST_VIN_3_G2)
    assert not multi_vehicle_controller.get_ev_status(TEST_VIN_4_SAFETY_PLUS)
    assert not multi_vehicle_controller.get_ev_status(TEST_VIN_5_G1_SECURITY)

    assert not multi_vehicle_controller.get_remote_status(TEST_VIN_1_G1)
    assert multi_vehicle_controller.get_remote_status(TEST_VIN_2_EV)
    assert multi_vehicle_controller.get_remote_status(TEST_VIN_3_G2)
    assert not multi_vehicle_controller.get_remote_status(TEST_VIN_4_SAFETY_PLUS)
    assert multi_vehicle_controller.get_remote_status(TEST_VIN_5_G1_SECURITY)

    assert not multi_vehicle_controller.get_res_status(TEST_VIN_1_G1)
    assert not multi_vehicle_controller.get_res_status(TEST_VIN_2_EV)
    assert multi_vehicle_controller.get_res_status(TEST_VIN_3_G2)
    assert not multi_vehicle_controller.get_res_status(TEST_VIN_4_SAFETY_PLUS)
    assert not multi_vehicle_controller.get_res_status(TEST_VIN_5_G1_SECURITY)


async def test_get_vehicle_status_ev_security_plus(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.get_data(TEST_VIN_2_EV.lower()))
    await add_validate_session(test_server)
    await add_select_vehicle_sequence(test_server, 2)
    await add_ev_vehicle_status(test_server)
    await add_validate_session(test_server)
    await add_ev_vehicle_condition(test_server)
    await add_validate_session(test_server)
    await add_g2_vehicle_locate(test_server)
    await add_validate_session(test_server)
    await add_vehicle_health(test_server)
    await add_fetch_climate_presets(test_server)
    status = (await task)[sc.VEHICLE_STATUS]
    assert status[sc.LOCATION_VALID]
    assert_vehicle_status(status, VEHICLE_STATUS_EV)
    assert_vehicle_condition(status, VEHICLE_CONDITION_EV)


async def test_get_vehicle_status_ev_bad_location(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.get_data(TEST_VIN_2_EV.lower()))
    await add_validate_session(test_server)
    await add_select_vehicle_sequence(test_server, 2)
    await add_ev_vehicle_status(test_server)
    await add_validate_session(test_server)
    await add_ev_vehicle_condition(test_server)
    await add_validate_session(test_server)
    await add_g2_vehicle_locate(test_server)
    await add_validate_session(test_server)
    await add_vehicle_health(test_server)
    await add_fetch_climate_presets(test_server)
    status = (await task)[sc.VEHICLE_STATUS]
    assert status[sc.LOCATION_VALID]
    assert_vehicle_status(status, VEHICLE_STATUS_EV)

    # Emulates a fetch after a Crosstrek PHEV is turned off, it will return bad coordinates
    task = asyncio.create_task(multi_vehicle_controller.fetch(TEST_VIN_2_EV.lower(), force=True))
    await add_validate_session(test_server)
    await add_ev_vehicle_status(test_server)
    await add_validate_session(test_server)
    await add_ev_vehicle_condition(test_server)
    await add_validate_session(test_server)
    await server_js_response(test_server, LOCATE_G2_BAD_LOCATION, path=API_LOCATE)
    await add_validate_session(test_server)
    await add_vehicle_health(test_server)
    await add_fetch_climate_presets(test_server)
    await task
    task = asyncio.create_task(multi_vehicle_controller.get_data(TEST_VIN_2_EV.lower()))
    status = (await task)[sc.VEHICLE_STATUS]

    # We should be informed that the current location is invalid/old
    assert not status[sc.LOCATION_VALID]

    # But still preserve the previous valid location
    assert_vehicle_status(status, VEHICLE_STATUS_EV)


async def test_get_vehicle_status_missing_data(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.get_data(TEST_VIN_4_SAFETY_PLUS))

    await add_validate_session(test_server)
    await add_select_vehicle_sequence(test_server, 4)

    # Manually set unreliable fields to good value
    good_data = VEHICLE_STATUS_EV["data"]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.TIRE_PRESSURE_FL] = good_data[
        API_TIRE_PRESSURE_FL
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.TIRE_PRESSURE_FR] = good_data[
        API_TIRE_PRESSURE_FR
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.TIRE_PRESSURE_RL] = good_data[
        API_TIRE_PRESSURE_RL
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.TIRE_PRESSURE_RR] = good_data[
        API_TIRE_PRESSURE_RR
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.AVG_FUEL_CONSUMPTION] = good_data[
        API_AVG_FUEL_CONSUMPTION
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.DIST_TO_EMPTY] = good_data[
        API_DIST_TO_EMPTY
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.LONGITUDE] = good_data[
        API_LONGITUDE
    ]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.LATITUDE] = good_data[API_LATITUDE]
    multi_vehicle_controller._vehicles[TEST_VIN_4_SAFETY_PLUS][sc.VEHICLE_STATUS][sc.VEHICLE_STATE] = good_data[
        API_VEHICLE_STATE
    ]

    # When VehicleStatus is missing data, controller should ignore and keep previous value
    await server_js_response(
        test_server,
        VEHICLE_STATUS_EV_MISSING_DATA,
        path=API_VEHICLE_STATUS,
    )
    status = (await task)[sc.VEHICLE_STATUS]
    assert_vehicle_status(status, VEHICLE_STATUS_EV)


async def test_update_g2(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.update(TEST_VIN_2_EV))

    await add_validate_session(test_server)
    await add_select_vehicle_sequence(test_server, 2)

    await server_js_response(
        test_server,
        VEHICLE_STATUS_EXECUTE,
        path=API_G2_LOCATE_UPDATE,
    )
    await server_js_response(test_server, VEHICLE_STATUS_STARTED, path=API_G2_LOCATE_STATUS)
    await server_js_response(
        test_server,
        VEHICLE_STATUS_FINISHED_SUCCESS,
        path=API_G2_LOCATE_STATUS,
    )

    assert await task


async def test_update_g1(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.update(TEST_VIN_5_G1_SECURITY))

    await add_validate_session(test_server)
    await server_js_response(
        test_server,
        LOCATE_G1_EXECUTE,
        path=API_G1_LOCATE_UPDATE,
    )
    await server_js_response(test_server, LOCATE_G1_STARTED, path=API_G1_LOCATE_STATUS)
    await server_js_response(
        test_server,
        LOCATE_G1_FINISHED,
        path=API_G1_LOCATE_STATUS,
    )

    assert await task



"""Tests for subarulink remote commands."""
import asyncio
import time

import pytest

from subarulink._subaru_api.const import (
    API_EV_CHARGE_NOW,
    API_G1_HORN_LIGHTS_STATUS,
    API_G2_FETCH_RES_SUBARU_PRESETS,
    API_G2_FETCH_RES_USER_PRESETS,
    API_G2_REMOTE_ENGINE_START,
    API_G2_REMOTE_ENGINE_STOP,
    API_G2_SAVE_RES_QUICK_START_SETTINGS,
    API_G2_SAVE_RES_SETTINGS,
    API_HORN_LIGHTS,
    API_HORN_LIGHTS_STOP,
    API_LIGHTS,
    API_LIGHTS_STOP,
    API_LOCK,
    API_LOGIN,
    API_REMOTE_SVC_STATUS,
    API_SELECT_VEHICLE,
    API_UNLOCK,
    API_VALIDATE_SESSION,
)
import subarulink.const as sc
from subarulink.exceptions import (
    InvalidPIN,
    PINLockoutProtect,
    RemoteServiceFailure,
    VehicleNotSupported,
)

from tests.api_responses import (
    FETCH_SUBARU_CLIMATE_PRESETS,
    FETCH_USER_CLIMATE_PRESETS_EV,
    LOCATE_G1_EXECUTE,
    LOCATE_G1_STARTED,
    LOGIN_MULTI_REGISTERED,
    REMOTE_CMD_INVALID_PIN,
    REMOTE_SERVICE_EXECUTE,
    REMOTE_SERVICE_STATUS_FINISHED_FAIL,
    REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
    REMOTE_SERVICE_STATUS_INVALID_TOKEN,
    REMOTE_SERVICE_STATUS_STARTED,
    SELECT_VEHICLE_2,
    SELECT_VEHICLE_3,
    SELECT_VEHICLE_5,
    SUBARU_PRESET_1,
    TEST_USER_PRESET_1,
    UPDATE_USER_CLIMATE_PRESETS,
    VALIDATE_SESSION_FAIL,
    VALIDATE_SESSION_SUCCESS,
)
from tests.conftest import (
    TEST_VIN_2_EV,
    TEST_VIN_3_G2,
    TEST_VIN_4_SAFETY_PLUS,
    TEST_VIN_5_G1_SECURITY,
    server_js_response,
)


async def test_remote_cmds_g2_ev(test_server, multi_vehicle_controller):
    cmd_list = [
        {
            "command": multi_vehicle_controller.horn(TEST_VIN_2_EV),
            "path": API_HORN_LIGHTS,
        },
        {
            "command": multi_vehicle_controller.horn_stop(TEST_VIN_2_EV),
            "path": API_HORN_LIGHTS_STOP,
        },
        {
            "command": multi_vehicle_controller.lights(TEST_VIN_2_EV),
            "path": API_LIGHTS,
        },
        {
            "command": multi_vehicle_controller.lights_stop(TEST_VIN_2_EV),
            "path": API_LIGHTS_STOP,
        },
        {
            "command": multi_vehicle_controller.lock(TEST_VIN_2_EV),
            "path": API_LOCK,
        },
        {
            "command": multi_vehicle_controller.unlock(TEST_VIN_2_EV),
            "path": API_UNLOCK,
        },
        {
            "command": multi_vehicle_controller.charge_start(TEST_VIN_2_EV),
            "path": API_EV_CHARGE_NOW,
        },
        {
            "command": multi_vehicle_controller.remote_stop(TEST_VIN_2_EV),
            "path": API_G2_REMOTE_ENGINE_STOP,
        },
    ]

    for cmd in cmd_list:
        task = asyncio.create_task(cmd["command"])
        await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)
        await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
        await server_js_response(
            test_server,
            SELECT_VEHICLE_2,
            path=API_SELECT_VEHICLE,
            query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
        )
        await server_js_response(
            test_server,
            REMOTE_SERVICE_EXECUTE,
            path=cmd["path"],
        )
        await server_js_response(
            test_server,
            REMOTE_SERVICE_STATUS_STARTED,
            path=API_REMOTE_SVC_STATUS,
        )
        await server_js_response(
            test_server,
            REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
            path=API_REMOTE_SVC_STATUS,
        )
        assert await task


async def test_remote_cmds_g1(test_server, multi_vehicle_controller):
    cmd_list = [
        {
            "command": multi_vehicle_controller.horn(TEST_VIN_5_G1_SECURITY),
            "path": API_HORN_LIGHTS,
            "status_url": API_G1_HORN_LIGHTS_STATUS,
        },
        {
            "command": multi_vehicle_controller.horn_stop(TEST_VIN_5_G1_SECURITY),
            "path": API_HORN_LIGHTS_STOP,
            "status_url": API_G1_HORN_LIGHTS_STATUS,
        },
        {
            "command": multi_vehicle_controller.lights(TEST_VIN_5_G1_SECURITY),
            "path": API_LIGHTS,
            "status_url": API_G1_HORN_LIGHTS_STATUS,
        },
        {
            "command": multi_vehicle_controller.lights_stop(TEST_VIN_5_G1_SECURITY),
            "path": API_LIGHTS_STOP,
            "status_url": API_G1_HORN_LIGHTS_STATUS,
        },
        {
            "command": multi_vehicle_controller.lock(TEST_VIN_5_G1_SECURITY),
            "path": API_LOCK,
            "status_url": API_REMOTE_SVC_STATUS,
        },
        {
            "command": multi_vehicle_controller.unlock(TEST_VIN_5_G1_SECURITY),
            "path": API_UNLOCK,
            "status_url": API_REMOTE_SVC_STATUS,
        },
    ]

    for cmd in cmd_list:
        task = asyncio.create_task(cmd["command"])
        await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)
        await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
        await server_js_response(
            test_server,
            SELECT_VEHICLE_5,
            path=API_SELECT_VEHICLE,
            query={"vin": TEST_VIN_5_G1_SECURITY, "_": str(int(time.time()))},
        )
        await server_js_response(
            test_server,
            REMOTE_SERVICE_EXECUTE,
            path=cmd["path"],
        )
        await server_js_response(test_server, REMOTE_SERVICE_STATUS_STARTED, path=cmd["status_url"])
        await server_js_response(
            test_server,
            REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
            path=cmd["status_url"],
        )
        assert await task


async def test_remote_cmds_unsupported(multi_vehicle_controller):
    cmd_list = [
        multi_vehicle_controller.horn(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.horn_stop(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.lights(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.lights_stop(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.lock(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.unlock(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.charge_start(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.remote_stop(TEST_VIN_4_SAFETY_PLUS),
        multi_vehicle_controller.remote_start(TEST_VIN_4_SAFETY_PLUS, "Test"),
        multi_vehicle_controller.update(TEST_VIN_4_SAFETY_PLUS),
    ]

    for cmd in cmd_list:
        task = asyncio.create_task(cmd)
        with pytest.raises(VehicleNotSupported):
            assert not await task


async def test_vehicle_remote_cmd_invalid_pin(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))

    assert not multi_vehicle_controller.invalid_pin_entered()
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(test_server, REMOTE_CMD_INVALID_PIN, path=API_LIGHTS)
    with pytest.raises(InvalidPIN):
        assert not await task
        assert multi_vehicle_controller.invalid_pin_entered()


async def test_vehicle_remote_cmd_invalid_pin_twice(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))

    assert not multi_vehicle_controller.invalid_pin_entered()
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_CMD_INVALID_PIN,
        path=API_LIGHTS,
    )
    with pytest.raises(InvalidPIN):
        assert not await task
        assert multi_vehicle_controller.invalid_pin_entered()

    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))
    with pytest.raises(PINLockoutProtect):
        await task


async def test_remote_cmd_failure(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_LIGHTS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_FAIL,
        path=API_REMOTE_SVC_STATUS,
    )
    with pytest.raises(RemoteServiceFailure):
        assert not await task


async def test_remote_cmd_timeout_g2(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_LIGHTS,
    )
    for _ in range(0, 20):
        await server_js_response(
            test_server,
            REMOTE_SERVICE_STATUS_STARTED,
            path=API_REMOTE_SVC_STATUS,
        )

    with pytest.raises(RemoteServiceFailure):
        assert not await task


async def test_remote_cmd_invalid_token(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_3_G2))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_LIGHTS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_INVALID_TOKEN,
        path=API_REMOTE_SVC_STATUS,
    )
    ## Session cookies cleared due to InvalidToken
    await server_js_response(test_server, VALIDATE_SESSION_FAIL, path=API_VALIDATE_SESSION)
    await server_js_response(test_server, LOGIN_MULTI_REGISTERED, path=API_LOGIN)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_3,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_3_G2, "_": str(int(time.time()))},
    )

    ## Continue with checking for remote service status
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=API_REMOTE_SVC_STATUS,
    )

    assert await task


async def test_remote_cmd_timeout_g1(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.lights(TEST_VIN_5_G1_SECURITY))

    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        LOCATE_G1_EXECUTE,
        path=API_LIGHTS,
    )
    for _ in range(0, 20):
        await server_js_response(test_server, LOCATE_G1_STARTED, path=API_G1_HORN_LIGHTS_STATUS)

    with pytest.raises(RemoteServiceFailure):
        assert not await task


async def test_delete_climate_preset_by_name(test_server, multi_vehicle_controller):
    task = asyncio.create_task(
        multi_vehicle_controller.delete_climate_preset_by_name(TEST_VIN_2_EV, TEST_USER_PRESET_1)
    )
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        UPDATE_USER_CLIMATE_PRESETS,
        path=API_G2_SAVE_RES_SETTINGS,
    )
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)
    assert await task


async def test_update_user_climate_presets(test_server, multi_vehicle_controller):
    new_preset_data = [
        {
            sc.REAR_AC: "false",
            sc.MODE: "AUTO",
            sc.FAN_SPEED: "AUTO",
            sc.TEMP_F: "71",
            sc.REAR_DEFROST: sc.REAR_DEFROST_OFF,
            sc.HEAT_SEAT_LEFT: "OFF",
            sc.HEAT_SEAT_RIGHT: "OFF",
            sc.RECIRCULATE: "outsideAir",
            sc.RUNTIME: "10",
            sc.PRESET_NAME: "Test",
        }
    ]

    task = asyncio.create_task(multi_vehicle_controller.update_user_climate_presets(TEST_VIN_2_EV, new_preset_data))
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )
    await server_js_response(
        test_server,
        UPDATE_USER_CLIMATE_PRESETS,
        path=API_G2_SAVE_RES_SETTINGS,
    )
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)
    assert await task


async def test_remote_start(test_server, multi_vehicle_controller):
    task = asyncio.create_task(multi_vehicle_controller.remote_start(TEST_VIN_2_EV, SUBARU_PRESET_1))
    await server_js_response(test_server, FETCH_SUBARU_CLIMATE_PRESETS, path=API_G2_FETCH_RES_SUBARU_PRESETS)
    await server_js_response(test_server, FETCH_USER_CLIMATE_PRESETS_EV, path=API_G2_FETCH_RES_USER_PRESETS)
    await server_js_response(
        test_server,
        UPDATE_USER_CLIMATE_PRESETS,
        path=API_G2_SAVE_RES_QUICK_START_SETTINGS,
    )
    await server_js_response(test_server, VALIDATE_SESSION_SUCCESS, path=API_VALIDATE_SESSION)
    await server_js_response(
        test_server,
        SELECT_VEHICLE_2,
        path=API_SELECT_VEHICLE,
        query={"vin": TEST_VIN_2_EV, "_": str(int(time.time()))},
    )

    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=API_G2_REMOTE_ENGINE_START,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_STARTED,
        path=API_REMOTE_SVC_STATUS,
    )
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=API_REMOTE_SVC_STATUS,
    )

    assert await task



#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

For more details, please refer to the documentation at https://github.com/G-Two/subarulink
"""

from subarulink import const
from subarulink.controller import Controller
from subarulink.exceptions import (
    IncompleteCredentials,
    InvalidCredentials,
    InvalidPIN,
    PINLockoutProtect,
    RemoteServiceFailure,
    SubaruException,
    VehicleNotSupported,
)

from .__version__ import __version__

__all__ = [
    "Controller",
    "SubaruException",
    "InvalidCredentials",
    "InvalidPIN",
    "IncompleteCredentials",
    "PINLockoutProtect",
    "RemoteServiceFailure",
    "VehicleNotSupported",
    "const",
    "__version__",
]

__pdoc__ = {"app": False}



#  SPDX-License-Identifier: Apache-2.0
"""
Provides exceptions specific to the subarulink package.

For more details, please refer to the documentation at https://github.com/G-Two/subarulink
"""


class SubaruException(Exception):
    """Class of Subaru API exceptions."""

    def __init__(self, message, *args, **kwargs):
        """Initialize exceptions for the Subaru Starlink API."""
        self.message = message
        super().__init__(*args, **kwargs)


class InvalidPIN(SubaruException):
    """Class of exceptions for invalid PIN number."""


class IncompleteCredentials(SubaruException):
    """Class of exceptions for not providing required credentials."""


class InvalidCredentials(SubaruException):
    """Class of exceptions for attempting to login with invalid credentials."""


class PINLockoutProtect(SubaruException):
    """Class of exception to notify previous invalid PIN is not being used to avoid account lockout."""


class VehicleNotSupported(SubaruException):
    """Class of exception when service requested is not supported by vehicle/subscription."""


class RemoteServiceFailure(SubaruException):
    """Class of exception when remote service fails."""



"""Tests for subarulink CLI."""
import asyncio
import sys
from unittest.mock import patch

import pytest

import subarulink._subaru_api.const as api
import subarulink.app.cli as cli

from tests.api_responses import (
    REMOTE_SERVICE_EXECUTE,
    REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
    REMOTE_SERVICE_STATUS_STARTED,
)
from tests.conftest import (
    TEST_COUNTRY,
    TEST_DEVICE_ID,
    TEST_PASSWORD,
    TEST_PIN,
    TEST_USERNAME,
    TEST_VIN_2_EV,
    add_ev_vehicle_condition,
    add_ev_vehicle_status,
    add_fetch_climate_presets,
    add_g2_vehicle_locate,
    add_multi_vehicle_login_sequence,
    add_select_vehicle_sequence,
    add_validate_session,
    add_vehicle_health,
    server_js_response,
)


@pytest.fixture
async def cli_controller(controller):
    instance = cli.CLI("tests/cli_test.cfg")
    instance.ctrl = controller
    instance.session = controller._connection._websession
    with patch("subarulink.app.cli.CLI._init_controller", return_value=None):
        yield instance


@pytest.fixture
async def interactive_session(test_server, cli_controller):
    input_values = ["2"]

    with patch("builtins.input", return_value=input_values.pop()):
        task = asyncio.create_task(cli_controller.run())
        await add_multi_vehicle_login_sequence(test_server)

        await add_validate_session(test_server)
        await add_select_vehicle_sequence(test_server, 2)

        await add_ev_vehicle_status(test_server)

        await add_validate_session(test_server)
        await add_ev_vehicle_condition(test_server)

        await add_validate_session(test_server)
        await add_g2_vehicle_locate(test_server)

        await add_validate_session(test_server)
        await add_vehicle_health(test_server)

        await add_fetch_climate_presets(test_server)

        yield task


def test_no_args():
    testargs = ["subarulink"]
    with patch.object(sys, "argv", testargs), patch("argparse.ArgumentParser.print_help") as mock_print_help:
        cli.main()
    assert mock_print_help.call_count == 1


def test_read_config():
    instance = cli.CLI("tests/cli_test.cfg")
    assert instance.config["country"] == TEST_COUNTRY
    assert instance.config["username"] == TEST_USERNAME
    assert instance.config["password"] == TEST_PASSWORD
    assert instance.config["pin"] == TEST_PIN
    assert instance.config["device_id"] == int(TEST_DEVICE_ID)


async def test_single_cmds(test_server, cli_controller):
    cmd_list = [
        {
            "command": "horn",
            "path": api.API_HORN_LIGHTS,
        },
        {
            "command": "lights",
            "path": api.API_LIGHTS,
        },
        {
            "command": "lock",
            "path": api.API_LOCK,
        },
        {
            "command": "unlock",
            "path": api.API_UNLOCK,
        },
        {
            "command": "charge",
            "path": api.API_EV_CHARGE_NOW,
        },
        {
            "command": "remote_stop",
            "path": api.API_G2_REMOTE_ENGINE_STOP,
        },
    ]
    cli_test = cli.CLI("tests/cli_test.cfg")
    for cmd in cmd_list:
        await run_single_cmd(test_server, cli_controller, cmd["command"], cmd["path"], cli_test.config)


async def run_single_cmd(test_server, cli_controller, cmd, path, config):
    task = asyncio.create_task(cli_controller.single_command(cmd, TEST_VIN_2_EV, config))

    await add_multi_vehicle_login_sequence(test_server)
    await add_validate_session(test_server)
    await add_select_vehicle_sequence(test_server, 2)
    await add_ev_vehicle_status(test_server)
    await add_validate_session(test_server)
    await add_ev_vehicle_condition(test_server)
    await add_validate_session(test_server)
    await add_g2_vehicle_locate(test_server)
    await add_validate_session(test_server)
    await add_vehicle_health(test_server)
    await add_fetch_climate_presets(test_server)
    await add_validate_session(test_server)

    await server_js_response(
        test_server,
        REMOTE_SERVICE_EXECUTE,
        path=path,
    )
    await server_js_response(test_server, REMOTE_SERVICE_STATUS_STARTED, path=api.API_REMOTE_SVC_STATUS)
    await server_js_response(
        test_server,
        REMOTE_SERVICE_STATUS_FINISHED_SUCCESS,
        path=api.API_REMOTE_SVC_STATUS,
    )

    with patch("sys.exit") as mock_exit:
        await task
    mock_exit.assert_called_once_with(0)


async def test_interactive_quit(interactive_session):
    with patch("builtins.input", return_value="quit"):
        await interactive_session


# TODO: More tests



#  SPDX-License-Identifier: Apache-2.0
"""
Constants for users of this package.

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""
from typing import Dict, Final, List

COUNTRY_USA = "USA"
COUNTRY_CAN = "CAN"

# Fields for users of this package (must stay consistent despite Subaru's changes)
# Home Assistant uses these names to establish unique_ids for entities
AVG_FUEL_CONSUMPTION = "AVG_FUEL_CONSUMPTION"
DIST_TO_EMPTY = "DISTANCE_TO_EMPTY_FUEL"
DOOR_BOOT_POSITION = "DOOR_BOOT_POSITION"
DOOR_ENGINE_HOOD_POSITION = "DOOR_ENGINE_HOOD_POSITION"
DOOR_FRONT_LEFT_POSITION = "DOOR_FRONT_LEFT_POSITION"
DOOR_FRONT_RIGHT_POSITION = "DOOR_FRONT_RIGHT_POSITION"
DOOR_REAR_LEFT_POSITION = "DOOR_REAR_LEFT_POSITION"
DOOR_REAR_RIGHT_POSITION = "DOOR_REAR_RIGHT_POSITION"
EV_CHARGER_STATE_TYPE = "EV_CHARGER_STATE_TYPE"
EV_DISTANCE_TO_EMPTY = "EV_DISTANCE_TO_EMPTY"
EV_IS_PLUGGED_IN = "EV_IS_PLUGGED_IN"
EV_STATE_OF_CHARGE_MODE = "EV_STATE_OF_CHARGE_MODE"
EV_STATE_OF_CHARGE_PERCENT = "EV_STATE_OF_CHARGE_PERCENT"
EV_TIME_TO_FULLY_CHARGED = "EV_TIME_TO_FULLY_CHARGED"
EV_TIME_TO_FULLY_CHARGED_UTC = "EV_TIME_TO_FULLY_CHARGED_UTC"
HEADING = "HEADING"
LAST_UPDATED_DATE = "LAST_UPDATED_DATE"
LATITUDE = "LATITUDE"
LONGITUDE = "LONGITUDE"
ODOMETER = "ODOMETER"
REMAINING_FUEL_PERCENT = "REMAINING_FUEL_PERCENT"
TIMESTAMP = "TIMESTAMP"
TIRE_PRESSURE_FL = "TYRE_PRESSURE_FRONT_LEFT"
TIRE_PRESSURE_FR = "TYRE_PRESSURE_FRONT_RIGHT"
TIRE_PRESSURE_RL = "TYRE_PRESSURE_REAR_LEFT"
TIRE_PRESSURE_RR = "TYRE_PRESSURE_REAR_RIGHT"
VEHICLE_STATE = "VEHICLE_STATE_TYPE"
WINDOW_FRONT_LEFT_STATUS = "WINDOW_FRONT_LEFT_STATUS"
WINDOW_FRONT_RIGHT_STATUS = "WINDOW_FRONT_RIGHT_STATUS"
WINDOW_REAR_LEFT_STATUS = "WINDOW_REAR_LEFT_STATUS"
WINDOW_REAR_RIGHT_STATUS = "WINDOW_REAR_RIGHT_STATUS"
WINDOW_SUNROOF_STATUS = "WINDOW_SUNROOF_STATUS"

# VEHICLE HEALTH STATUS
HEALTH_FEATURES = "FEATURES"
HEALTH_MIL_NAME = "MIL"
HEALTH_TROUBLE = "ISTROUBLE"
HEALTH_ONDATE = "ONDATE"
HEALTH_RECOMMENDED_TIRE_PRESSURE = "RECOMMENDED_TIRE_PRESSURE"
HEALTH_RECOMMENDED_TIRE_PRESSURE_FRONT = "FRONT_TIRES"
HEALTH_RECOMMENDED_TIRE_PRESSURE_REAR = "REAR_TIRES"

# EV_IS_PLUGGED_IN Values
CHARGING = "CHARGING"
LOCKED_CONNECTED = "LOCKED_CONNECTED"
UNLOCKED_CONNECTED = "UNLOCKED_CONNECTED"

# Door Values
DOOR_OPEN = "OPEN"
DOOR_CLOSED = "CLOSED"

# Window Values
WINDOW_OPEN = "OPEN"
WINDOW_VENTED = "VENTED"
WINDOW_CLOSED = "CLOSE"

# Moonroof Values
SUNROOF_OPEN = "OPEN"
SUNROOF_SLIDE_PARTLY_OPEN = "SLIDE_PARTLY_OPEN"
SUNROOF_TILT = "TILT"
SUNROOF_TILT_PARTLY_OPEN = "TILT_PARTLY_OPEN"
SUNROOF_CLOSED = "CLOSE"

# VEHICLE_STATE Values
IGNITION_ON = "IGNITION_ON"
IGNITION_OFF = "IGNITION_OFF"

NOT_EQUIPPED = "NOT_EQUIPPED"

# Remote start constants
TEMP_F = "climateZoneFrontTemp"
TEMP_F_MAX = 85
TEMP_F_MIN = 60

TEMP_C = "climateZoneFrontTempCelsius"
TEMP_C_MAX = 30
TEMP_C_MIN = 15

RUNTIME = "runTimeMinutes"
RUNTIME_10_MIN = "10"
RUNTIME_5_MIN = "5"

MODE = "climateZoneFrontAirMode"
MODE_DEFROST = "WINDOW"
MODE_FEET_DEFROST = "FEET_WINDOW"
MODE_FACE = "FACE"
MODE_FEET = "FEET"
MODE_SPLIT = "FEET_FACE_BALANCED"
MODE_AUTO = "AUTO"

HEAT_SEAT_LEFT = "heatedSeatFrontLeft"
HEAT_SEAT_RIGHT = "heatedSeatFrontRight"
HEAT_SEAT_HI = "HIGH_HEAT"
HEAT_SEAT_MED = "MEDIUM_HEAT"
HEAT_SEAT_LOW = "LOW_HEAT"
HEAT_SEAT_OFF = "OFF"

REAR_DEFROST = "heatedRearWindowActive"
REAR_DEFROST_ON = "true"
REAR_DEFROST_OFF = "false"

FAN_SPEED = "climateZoneFrontAirVolume"
FAN_SPEED_LOW = "2"
FAN_SPEED_MED = "4"
FAN_SPEED_HI = "7"
FAN_SPEED_AUTO = "AUTO"

RECIRCULATE = "outerAirCirculation"
RECIRCULATE_OFF = "outsideAir"
RECIRCULATE_ON = "recirculation"

REAR_AC = "airConditionOn"
REAR_AC_ON = "true"
REAR_AC_OFF = "false"

PRESET_NAME = "name"
PRESET_INDEX = "index"

CAN_EDIT = "canEdit"
CAN_EDIT_VALUE = "true"
DISABLED = "disabled"
DISABLED_VALUE = "false"
PRESET_TYPE = "presetType"
PRESET_TYPE_USER = "userPreset"
START_CONFIGURATION = "startConfiguration"
START_CONFIGURATION_EV = "START_CLIMATE_CONTROL_ONLY_ALLOW_KEY_IN_IGNITION"
START_CONFIGURATION_RES = "START_ENGINE_ALLOW_KEY_IN_IGNITION"

START_CONFIG_CONSTS_EV = {
    CAN_EDIT: CAN_EDIT_VALUE,
    DISABLED: DISABLED_VALUE,
    PRESET_TYPE: PRESET_TYPE_USER,
    START_CONFIGURATION: START_CONFIGURATION_EV,
}
START_CONFIG_CONSTS_RES = {
    CAN_EDIT: CAN_EDIT_VALUE,
    DISABLED: DISABLED_VALUE,
    PRESET_TYPE: PRESET_TYPE_USER,
    START_CONFIGURATION: START_CONFIGURATION_RES,
}

VALID_CLIMATE_OPTIONS: Dict[str, List] = {
    TEMP_C: [str(_) for _ in range(TEMP_C_MIN, TEMP_C_MAX + 1)],
    TEMP_F: [str(_) for _ in range(TEMP_F_MIN, TEMP_F_MAX + 1)],
    FAN_SPEED: [FAN_SPEED_AUTO, FAN_SPEED_LOW, FAN_SPEED_MED, FAN_SPEED_HI],
    HEAT_SEAT_LEFT: [HEAT_SEAT_OFF, HEAT_SEAT_LOW, HEAT_SEAT_MED, HEAT_SEAT_HI],
    HEAT_SEAT_RIGHT: [HEAT_SEAT_OFF, HEAT_SEAT_LOW, HEAT_SEAT_MED, HEAT_SEAT_HI],
    MODE: [
        MODE_DEFROST,
        MODE_FEET_DEFROST,
        MODE_FACE,
        MODE_FEET,
        MODE_SPLIT,
        MODE_AUTO,
    ],
    RECIRCULATE: [RECIRCULATE_OFF, RECIRCULATE_ON],
    REAR_AC: [REAR_AC_OFF, REAR_AC_ON],
    REAR_DEFROST: [REAR_DEFROST_OFF, REAR_DEFROST_ON],
    RUNTIME: [RUNTIME_10_MIN, RUNTIME_5_MIN],
    PRESET_NAME: [None],
    CAN_EDIT: [CAN_EDIT_VALUE],
    DISABLED: [DISABLED_VALUE],
    PRESET_TYPE: [PRESET_TYPE_USER],
    START_CONFIGURATION: [START_CONFIGURATION_EV, START_CONFIGURATION_RES],
    PRESET_INDEX: [0, 1, 2, 3],
}

# Unlock doors constants
WHICH_DOOR = "unlockDoorType"
ALL_DOORS = "ALL_DOORS_CMD"
DRIVERS_DOOR = "FRONT_LEFT_DOOR_CMD"
TAILGATE_DOOR = "TAILGATE_DOOR_CMD"
VALID_DOORS = [ALL_DOORS, DRIVERS_DOOR, TAILGATE_DOOR]


# Erroneous Values
BAD_AVG_FUEL_CONSUMPTION = "16383"
BAD_DISTANCE_TO_EMPTY_FUEL = "16383"
BAD_EV_TIME_TO_FULLY_CHARGED = "65535"
BAD_TIRE_PRESSURE = "32767"
BAD_LONGITUDE = 180
BAD_LATITUDE = 90
BAD_ODOMETER = None
BAD_SENSOR_VALUES = [
    BAD_AVG_FUEL_CONSUMPTION,
    BAD_DISTANCE_TO_EMPTY_FUEL,
    BAD_EV_TIME_TO_FULLY_CHARGED,
    BAD_TIRE_PRESSURE,
    BAD_ODOMETER,
]
UNKNOWN = "UNKNOWN"
BAD_BINARY_SENSOR_VALUES = [UNKNOWN, NOT_EQUIPPED]

LOCATION_VALID = "LOCATION_VALID"

FEATURE_G1_TELEMATICS = "g1"
FEATURE_G2_TELEMATICS = "g2"
FEATURE_G3_TELEMATICS = "g3"

POLL_INTERVAL = 7200
FETCH_INTERVAL = 300

VEHICLE_INFO: Final = "vehicle_info"
VEHICLE_STATUS: Final = "vehicle_status"
VEHICLE_HEALTH: Final = "vehicle_health"
VEHICLE_MODEL_YEAR: Final = "model_year"
VEHICLE_MODEL_NAME: Final = "model_name"
VEHICLE_NAME: Final = "vehicle_name"
VEHICLE_FEATURES: Final = "vehicle_features"
VEHICLE_SUBSCRIPTION_FEATURES: Final = "subscription_features"
VEHICLE_SUBSCRIPTION_STATUS: Final = "subscription_status"
VEHICLE_CLIMATE: Final = "climate"
VEHICLE_LAST_FETCH: Final = "last_fetch"
VEHICLE_LAST_UPDATE: Final = "last_update"


# List of potentially sensitive or identifying data fields to redact from raw API data when using HA diagnostic feature
RAW_API_FIELDS_TO_REDACT = [
    "cachedStateCode",
    "customer",
    "email",
    "firstName",
    "lastName",
    "latitude",
    "licensePlate",
    "licensePlateState",
    "longitude",
    "nickname",
    "odometer",
    "odometerValue",
    "odometerValueKilometers",
    "oemCustId",
    "phone",
    "preferredDealer",
    "sessionCustomer",
    "timeZone",
    "userOemCustId",
    "vehicleGeoPosition",
    "vehicleKey",
    "vehicleMileage",
    "vehicleName",
    "vhsId",
    "vin",
    "zip",
]
